/*
 * Copyright (C) 2003 BLADOX, s.r.o.  All rights reserved.
 *
 * This file is part of an example program for Turbo. This example
 * program may be used, distributed and modified without limitation.
 *
 */

#include <config.h>
#include <tprog/tprog.h>

#include "i2cma.h"
#include "brum2i2c.h"

u8 sladdr;

void init_brum2i2c (u8 myslaveaddr)
{
  brum2i2c_dbgs ("init_brum2i2c\n");
  i2c_init (myslaveaddr);
}

void brum2i2c_selectSlave (u8 slaveaddr)
{
  brum2i2c_dbgs ("brum2i2c_selectSlave\n");
  sladdr = slaveaddr;
}

void brum2i2c_getSlave (u8 * slaveaddr)
{
  brum2i2c_dbgs ("brum2i2c_getSlave\n");
  *slaveaddr = sladdr;
}

// TRANSMIT function
u8 brum2i2c_tx_data (u8 * data, u8 len)
{
  u8 status;
  u8 i;
  u8 reps;

  brum2i2c_dbgs ("brum2i2c_tx_data\n");
  reps = 0;
I2C_SEND_DATA_ERROR_TX_REP:
  status = i2c_send_start ();	//send START
  if (status != 0x08)
    goto I2C_SEND_DATA_ERROR_TX;	//finish with error code
  status = i2c_send_sla_w (sladdr);	//send SLA+W
  if (status != 0x18)
    goto I2C_SEND_DATA_ERROR_TX;	//finish with error code
  for (i = 0; i < len; i++)
  {
    brum2i2c_dbgc (data[i]);
    status = i2c_send_data (data[i]);	//send DATA
    if (status != 0x28)
      goto I2C_SEND_DATA_ERROR_TX;	//finish with error code
  }
  status = i2c_send_stop ();	//send STOP
  return 0;
I2C_SEND_DATA_ERROR_TX:
  if (status == 0x20)
  {
    delayUs (100);
    reps++;
    if (reps != 100)
    {
      status = i2c_send_stop ();	//send STOP
      goto I2C_SEND_DATA_ERROR_TX_REP;	//finish with error code
    }
  }
  brum2i2c_dbgs ("brum2i2c_tx_data\n");
  brum2i2c_dbgs ("err=");
  brum2i2c_dbgs ("0x");
  brum2i2c_dbgch (status);
  brum2i2c_dbgs ("\n");
  return status;
}

// RECEIVE function
u8 brum2i2c_rx_data (u8 * data, u8 len)
{
  u8 status;
  u8 i;
  u8 reps;

  brum2i2c_dbgs ("brum2i2c_rx_data\n");
  reps = 0;
I2C_SEND_DATA_ERROR_RX_REP:
  status = i2c_send_start ();	//send START
  if (status != 0x08)
    goto I2C_SEND_DATA_ERROR_RX;	//finish with error code
  status = i2c_send_sla_r (sladdr);	//send SLA+R
  if (status != 0x40)
    goto I2C_SEND_DATA_ERROR_RX;	//finish with error code
  for (i = 0; i < len; i++)
  {
    if (i != (len - 1))
    {				//not last byte te read, generate ACK
      status = i2c_receive_data ((u8 *) & (data[i]), 0);	//receive not last DATA
      brum2i2c_dbgc (data[i]);
      if (status != 0x50)
	goto I2C_SEND_DATA_ERROR_RX;	//finish with error code
    }
    else
    {				//last byte te read, generate NACK
      brum2i2c_dbgc (data[i]);
      status = i2c_receive_data (&(data[i]), 1);	//receive last DATA
      if (status != 0x58)
	goto I2C_SEND_DATA_ERROR_RX;	//finish with error code
    }
  }
  status = i2c_send_stop ();	//send STOP
  return 0;
I2C_SEND_DATA_ERROR_RX:
  if (status == 0x48)
  {
    delayUs (100);
    reps++;
    if (reps != 100)
    {
      status = i2c_send_stop ();	//send STOP
      goto I2C_SEND_DATA_ERROR_RX_REP;	//finish with error code
    }
  }
  brum2i2c_dbgs ("rx_data\n");
  brum2i2c_dbgs ("err=");
  brum2i2c_dbgs ("0x");
  brum2i2c_dbgch (status);
  brum2i2c_dbgs ("\n");
  return status;
}

// SELECT register
u8 brum2i2c_select_reg (u8 reg)
{
  u8 buf[1];
  u8 status;

  brum2i2c_dbgs ("brum2i2c_select_reg");
  brum2i2c_dbgch (reg);
  brum2i2c_dbgs ("\n");
  buf[0] = reg;			//pointer reg.
  status = brum2i2c_tx_data (buf, 1);
  return status;
}

// READ rxfifo data
u8 brum2i2c_read_rxfifo_data (u8 * outbuf, u8 len)
{
  u8 status;
  u8 i;

//__dbs("urrd");__dbs("\n");
  brum2i2c_select_reg (BRUMMODULE_RXFIFO_DATA);
  status = brum2i2c_rx_data (outbuf, len);
//for(i=0;i<len;i++) {__dbc(outbuf[i]); }
  return status;
}

// WRITE txfifo data
u8 brum2i2c_write_txfifo_data (u8 * inbuf, u8 len)
{
  u8 buf[64];
  u8 status;
  u8 i;

//__dbs("uwtd");__dbs("\n");
  buf[0] = BRUMMODULE_TXFIFO_DATA;	//pointer reg.
  for (i = 0; i < len; i++)
  {
    buf[i + 1] = inbuf[i];
  }
//for(i=0;i<len;i++) { __dbc(inbuf[i]);}
  status = brum2i2c_tx_data (buf, len + 1);
  return status;
}

// READ rxfifo status
u8 brum2i2c_read_rxfifo_status (u8 * val)
{
  u8 status;
  u8 i;

//__dbs("urrs");__dbs("\n");
  brum2i2c_select_reg (BRUMMODULE_RXFIFO_STATUS);
  status = brum2i2c_rx_data (val, 1);
  return status;
}

// READ txfifo status
u8 brum2i2c_read_txfifo_status (u8 * val)
{
  u8 status;
  u8 i;

//__dbs("urts");__dbs("\n");
  brum2i2c_select_reg (BRUMMODULE_TXFIFO_STATUS);
  status = brum2i2c_rx_data (val, 1);
  return status;
}

// READ rxfifo command
u8 brum2i2c_read_rxfifo_cmd (u8 * val)
{
  u8 status;
  u8 i;

//__dbs("urrc");__dbs("\n");
  brum2i2c_select_reg (BRUMMODULE_RXFIFO_CMD);
  status = brum2i2c_rx_data (val, 1);
  return status;
}

// WRITE rxfifo command
u8 brum2i2c_write_rxfifo_cmd (u8 val)
{
  u8 buf[2];
  u8 status;

//__dbs("uwrc");__dbs("\n");
  buf[0] = BRUMMODULE_RXFIFO_CMD;	//pointer reg.
  buf[1] = val;
  status = brum2i2c_tx_data (buf, 2);
  return status;
}

// READ txfifo command
u8 brum2i2c_read_txfifo_cmd (u8 * val)
{
  u8 status;
  u8 i;

//__dbs("urtc");__dbs("\n");
  brum2i2c_select_reg (BRUMMODULE_TXFIFO_CMD);
  status = brum2i2c_rx_data (val, 1);
  return status;
}

// WRITE txfifo command
u8 brum2i2c_write_txfifo_cmd (u8 val)
{
  u8 buf[2];
  u8 status;

//__dbs("uwtc");__dbs("\n");
  buf[0] = BRUMMODULE_TXFIFO_CMD;	//pointer reg.
  buf[1] = val;
  status = brum2i2c_tx_data (buf, 2);
  return status;
}

// READ control
u8 brum2i2c_read_control (u8 * val)
{
  u8 status;
  u8 i;

//__dbs("uruh");__dbs("\n");
  brum2i2c_select_reg (BRUMMODULE_STATUS_CONTROL);
  status = brum2i2c_rx_data (val, 1);
  return status;
}

// WRITE control
u8 brum2i2c_write_control (u8 val)
{
  u8 buf[2];
  u8 status;

//__dbs("uwuh");__dbs("\n");
  buf[0] = BRUMMODULE_STATUS_CONTROL;	//pointer reg.
  buf[1] = val;
  status = brum2i2c_tx_data (buf, 2);
  return status;
}

// READ rxchars
u8 brum2i2c_read_rxchars (u16 * val)
{
  u8 buf[2];
  u8 status;
  u8 i;

//__dbs("urrl");__dbs("\n");
  brum2i2c_select_reg (BRUMMODULE_RXCHARS);
  status = brum2i2c_rx_data (buf, 2);
  *val = (buf[1] << 8) | buf[0];
  return status;
}

// READ txchars
u8 brum2i2c_read_txchars (u16 * val)
{
  u8 buf[2];
  u8 status;
  u8 i;

//__dbs("urtl");__dbs("\n");
  brum2i2c_select_reg (BRUMMODULE_TXCHARS);
  status = brum2i2c_rx_data (buf, 2);
  *val = (buf[1] << 8) | buf[0];
  return status;
}

// READ 
u8 brum2i2c_read_debug_chars (u16 * val)
{
  u8 buf[2];
  u8 status;
  u8 i;

  brum2i2c_select_reg (BRUMMODULE_DEBUGFIFO_CHARS);
  status = brum2i2c_rx_data (buf, 2);
  *val = (buf[1] << 8) | buf[0];
  return status;
}

// READ 
u8 brum2i2c_read_debug_data (u8 * outbuf, u8 len)
{
  u8 status;
  u8 i;

  brum2i2c_select_reg (BRUMMODULE_DEBUGFIFO_DATA);
  status = brum2i2c_rx_data (outbuf, len);
  return status;
}

// READ rssi
u8 brum2i2c_read_rssi (u8 * val)
{
  u8 status;
  u8 i;

//__dbs("urr");__dbs("\n");
  brum2i2c_select_reg (BRUMMODULE_CC_RSSI);
  status = brum2i2c_rx_data (val, 1);
  return status;
}

// READ id
u8 brum2i2c_read_id (u8 * val)
{
  u8 status;
  u8 i;

//__dbs("uri");__dbs("\n");
  brum2i2c_select_reg (BRUMMODULE_ID);
  status = brum2i2c_rx_data (val, 1);
  return status;
}
