/*
 * Turbo Programmer Utilities, turbo-prog-utils, www.bladox.com 
 *
 * Copyright (C) 2004 BLADOX, s.r.o.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <tproglib.h>
#include <stdlib.h>

#include "../module-src/df.h"

u8 dataFlashStatusRegisterRead ()
{
  u8 answer[2048];
  u16 answerLen;

  usb_send (ACTION_DF_READ_STATUS, 0, NULL, &answerLen, answer);

  return answer[0];
}

u8 dataFlashBufferRead (u8 bufferNumber, u8 * buffer, u16 addr, u16 size)
{
  u8 txBuf[2048];
  u16 answerLen;

  txBuf[0] = bufferNumber;	//BUFFER NUM
  txBuf[1] = addr;		//BUFFER OFFSET low byte
  txBuf[2] = addr >> 8;		//BUFFER OFFSET high byte
  txBuf[3] = (size & 0xFF);	//NUM OF BYTES TO READ low byte
  txBuf[4] = (size >> 8);	//NUM OF BYTES TO READ high byte
  usb_send (ACTION_DF_BUFFER_READ, 5, txBuf, &answerLen, buffer);
}

u8 dataFlashMemoryToBufferRead (u8 bufferNumber, u16 pageAddr)
{
  u8 txBuf[2048];
  u8 answer[2048];
  u16 answerLen;

  txBuf[0] = bufferNumber;	//BUFFER NUM
  txBuf[1] = (pageAddr & 0xFF);	//PAGE NUM low byte
  txBuf[2] = (pageAddr >> 8);	//PAGE NUM high byte

  usb_send (ACTION_DF_MEMORY_TO_BUFFER_READ, 3, txBuf, &answerLen, answer);

  return 0;
}

u8 dataFlashBufferWrite (u8 bufferNumber, u8 * buffer, u16 addr, u16 pageSize)
{
  u16 i;
  u8 txBuf[2048];
  u8 answer[2048];
  u16 answerLen;

  txBuf[0] = bufferNumber;	//BUFFER NUM
  txBuf[1] = addr;		//BUFFER OFFSET low byte
  txBuf[2] = addr >> 8;		//BUFFER OFFSET high byte
  txBuf[3] = (pageSize & 0xFF);	//NUM OF BYTES TO READ low byte
  txBuf[4] = (pageSize >> 8);	//NUM OF BYTES TO READ high byte
  for (i = 0; i < pageSize; i++)
    txBuf[5 + i] = buffer[i];

  usb_send (ACTION_DF_BUFFER_WRITE, 5 + pageSize, txBuf, &answerLen, answer);

  return 0;
}

u8 dataFlashBufferToMemoryProgramWithErase (u8 bufferNumber, u16 pageNumber)
{
  u8 txBuf[2048];
  u8 answer[2048];
  u16 answerLen;

  txBuf[0] = bufferNumber;	//BUFFER NUM
  txBuf[1] = (pageNumber & 0xFF);	//PAGE NUM low byte
  txBuf[2] = (pageNumber >> 8);	//PAGE NUM high byte

  usb_send (ACTION_DF_BUFFER_TO_MEMORY_PROGRAM_WITH_ERASE, 3, txBuf,
	    &answerLen, answer);

  return 0;
}

u8 dataFlashBlockErase (u16 pageNumber)
{
  u8 txBuf[2048];
  u8 answer[2048];
  u16 answerLen;

  txBuf[0] = (pageNumber & 0xFF);	//low byte
  txBuf[1] = (pageNumber >> 8);	//high byte

  usb_send (ACTION_DF_ERASE_BLOCK, 3, txBuf, &answerLen, answer);

  return 0;
}

u8 dataFlashPageErase (u16 pageNumber)
{
  u8 txBuf[2048];
  u8 answer[2048];
  u16 answerLen;
  u8 i;

  txBuf[0] = (pageNumber & 0xFF);	//low byte
  txBuf[1] = (pageNumber >> 8);	//high byte
  usb_send (ACTION_DF_ERASE_PAGE, 2, txBuf, &answerLen, answer);

  return 0;
}
