/*
 * Turbo Programmer Utilities, turbo-prog-utils, www.bladox.com 
 *
 * Copyright (C) 2004 BLADOX, s.r.o.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <tfslib.h>
#include <tproglib.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include "../module-src/df.h"

void help ()
{
  printf ("Encrypt/Decrypt file on card\n");
  printf ("Usage: [args] inode_in_hex\n");
  printf ("-h\thelp\n");
  printf ("-d device\tuse USB device\n");
  printf ("-e\tencrypt\n");
  printf ("-7\t7bit alphabet (packed)\n");
  printf ("-p <password>\tpassword\n");
  printf ("-s\tstrong (also len and magic)\n");
  printf ("-v\tverbose\n");
  exit (-1);
}

int main (int argc, char **argv)
{
  u32 i;
  u16 c;
  u8 nb;
  u16 pos;
  u8 buf[1000];
  u8 len;
  int result;
  u8 sort;
  u16 j;
  u8 *d_name;
  u8 path[1000] = { '/', '\0' };
  u8 name[1000];
  u16 name_end;
  u16 path_end;
  b32 f;
  Inode ino;
  u8 enc = 0;
  u8 *pswd = NULL;
  u8 key[16];
  u8 strong = 0;
  u8 b7 = 0;
  u8 *dev = USBDEVICE;

  do
  {
    result = getopt (argc, argv, "ep:7shd:v");
    switch (result)
    {
      case 'd':
	dev = optarg;
	break;
      case 'e':
	enc = 1;
	break;
      case 's':
	strong = 1;
	break;
      case '7':
	b7 = 1;
	break;
      case 'p':
	pswd = optarg;
	break;
      case 'h':
	help ();
	break;
      case 'v':
	verbose_inc ();
	break;

    }
  }
  while (result > 0);

  if (pswd == NULL)
  {
    fprintf (stderr, "Missing password\n");
    help ();
  }

  strncpy (key, pswd, 16);

  if (argc - optind != 1)
    help ();
  sscanf (argv[optind], "%x", &i);
  f = fs_ino_p (i >> 8, i & 0xFF);
//printf("DST %x %04X %02X\n", f, fs_ino_block(f), fs_ino_index(f));

  open_device (dev);
  if (select_app (APP_DF) != NO_ERROR)
  {
    close_device ();
    fprintf (stderr, "Cannot open application %s\n", APP_DF);
    exit (-1);
  }

  usb_send (ACTION_DF_INIT, 0, NULL, NULL, NULL);

  if (mc_detect () == MC_OK)
  {
    u8 name[100];

    printf ("Memory card detected\n");
    mc_get_label (name);
    printf ("Card label: %s\n\n", name);
  }
  else if (mc_detect () == MC_NOT_PRESENT)
  {
    printf ("Memory card NOT PRESENT\n");
    close_device ();
    exit (1);
  }
  else if (mc_detect () == MC_NOT_READY)
  {
    printf ("Memory card NOT FORMATED\n\n");
    exit (1);
  }

  get_inode (&ino, f);
  printf ("Inode: %04X%02X\n", fs_ino_block (f), fs_ino_index (f));

  if (enc == 1 && ino.type != INODE_TYPE_FILE)
  {
    fprintf (stderr, "Not a regular file\n");
    goto X1;
  }
  if (enc == 0
      && (ino.type != INODE_TYPE_PROTECTED_KEY_DATA
	  && ino.type != INODE_TYPE_PROTECTED_DATA_LEN_TYPE))
  {
    fprintf (stderr, "Not encrypted file\n");
    goto X1;
  }

  if (enc == 0 && ino.type == INODE_TYPE_PROTECTED_DATA_LEN_TYPE)
    strong = 1;

  if (enc == 0)
  {
    //decrypt
    nb = mc_file_decrypt (f, key);
    if (nb == ERR_WRONG_KEY)
    {
      fprintf (stderr, "Wrong key\n");
      goto X1;
    }
    if (nb != NO_ERROR)
    {
      fprintf (stderr, "ERROR %d\n", nb);
      goto X1;
    }
  }
  else
  {
    //encrypt
    if (strong)
      mc_file_encrypt (f, key, INODE_TYPE_PROTECTED_DATA_LEN_TYPE);
    else
      mc_file_encrypt (f, key, INODE_TYPE_PROTECTED_KEY_DATA);
  }

X1:
  close_device ();
}
