/*
 * Turbo Programmer Utilities, turbo-prog-utils, www.bladox.com 
 *
 * Copyright (C) 2004 BLADOX, s.r.o.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <tfslib.h>
#include <tproglib.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include "../module-src/df.h"

/*
TODO
test if name exists -> do not create
*/

void help ()
{
  printf ("Make directory\n");
  printf ("Usage: [args] dir_name\n");
  printf ("-h\thelp\n");
  printf ("-d device\tuse USB device\n");
  printf ("-sx\tsort x (0=append, 1=sms, 2=sms_skip_2, 3=sort)\n");
  printf ("-v\tverbose\n");
  exit (-1);
}

int main (int argc, char **argv)
{
  u8 i;
  u16 c;
  u8 nb;
  u16 pos;
  u8 buf[1000];
  u8 len;
  int result;
  u8 sort;
  u16 j;
  u8 *d_name;
  u8 path[1000] = { '/', '\0' };
  u8 name[1000];
  u16 name_end;
  u16 path_end;
  b32 f;
  u8 *dev = USBDEVICE;

  do
  {
    result = getopt (argc, argv, "s:hd:v");
    switch (result)
    {
      case 'd':
	dev = optarg;
	break;
      case 's':
	sort = atoi (optarg);
	break;
      case 'h':
	help ();
	break;
      case 'v':
	verbose_inc ();
	break;
    }
  }
  while (result > 0);

  if (argc - optind != 1)
    help ();
  d_name = argv[optind++];
  printf ("DST %s\n", d_name);

  name_end = strlen (d_name);

  for (i = name_end; i > 0; i--)
    if (d_name[i - 1] == '/')
      break;
  path_end = i;

  len = strlen (d_name);
  for (j = 0; j < name_end - path_end; j++)
    name[j] = d_name[i + j];
  name[j] = '\0';

  for (j = 0; j < path_end; j++)
    path[j] = d_name[j];

  printf ("NAME %s\n", name);
  printf ("PATH %s\n", path);

  open_device (dev);
  if (select_app (APP_DF) != NO_ERROR)
  {
    close_device ();
    fprintf (stderr, "Cannot open application %s\n", APP_DF);
    exit (-1);
  }

  usb_send (ACTION_DF_INIT, 0, NULL, NULL, NULL);

  if (mc_detect () == MC_OK)
  {
    u8 name[100];

    printf ("Memory card detected\n");
    mc_get_label (name);
    printf ("Card label: %s\n\n", name);
  }
  else if (mc_detect () == MC_NOT_PRESENT)
  {
    printf ("Memory card NOT PRESENT\n");
    close_device ();
    exit (1);
  }
  else if (mc_detect () == MC_NOT_READY)
  {
    printf ("Memory card NOT FORMATED\n\n");
    exit (1);
  }

  f = cd (path);
  if (f < 0)
  {
    fprintf (stderr, "Path does not exist\n");
    exit (1);
  }
  f = create_dir (f, name, sort);
  if (f < 0)
  {
    fprintf (stderr, "ERROR %d\n", f);
    exit (1);
  }
  printf ("RESULT %04X%02X\n", fs_ino_block (f), fs_ino_index (f));
  close_device ();
}
