/*****************************************************************
/
/ File   :   ifdhandler.c
/ Author :   David Corcoran <corcoran@linuxnet.com>
/ Date   :   June 15, 2000
/ Purpose:   This provides reader specific low-level calls.
/            See http://www.linuxnet.com for more information.
/ License:   See file LICENSE
/
******************************************************************/

//#define IFDHANDLERv2
//#include <pcscdefines.h>
//#include <pcsclite.h>
#include <PCSC/ifdhandler.h>
#include <PCSC/debuglog.h>

#define MY_DBG	PCSC_LOG_INFO

#define MY_LUN_SIM	0
#define MY_LUN_TWIN	1

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/stat.h>

#include <tfslib.h>
#include <tproglib.h>
#include <cable.h>

#include "../apps/sim/module-src/sim.h"


u8 my_atr[1000];
u16 my_atr_len=0;

u8 my_atr_twin[1000];
u8 my_atr_twin_len=0;

u8 my_init=0;
u8 my_dev[1000];

u8 ef_sms_path[] = { 0x3f, 0x00, 0x7f, 0x10, 0x6f, 0x3c };

#define TO_SIM          0
#define FROM_SIM        1

u16 sim (u8 ins, u8 p1, u8 p2, u8 p3, u8 * buf)
{
  u16 len = p3;
  u16 i;
  u8 dir = TO_SIM;
  u16 ret;
  u8 txBuf[2048];

  switch (ins)
  {
//FROM_SIM
    case ME_CMD_STATUS:
    case ME_CMD_READ_BINARY:
   case ME_CMD_READ_RECORD:
    case ME_CMD_GET_RESPONSE:
    case ME_CMD_FETCH:
      dir = FROM_SIM;
      break;
    default:
      break;
/*
        case ME_CMD_INVALIDATE:
        case ME_CMD_REHABILITATE:
        case ME_CMD_SLEEP:
*/
  }

  if (len == 0 && !(ins == ME_CMD_INVALIDATE || ins == ME_CMD_REHABILITATE))
    len = 0x100;
  if (dir == FROM_SIM)
   len = 0;

  //GSM command
  txBuf[0] = ins;
  txBuf[1] = p1;
  txBuf[2] = p2;
  txBuf[3] = p3;

  for (i = 0; i < len; i++)
    txBuf[4 + i] = buf[i];

  usb_send (ACTION_SIM_CMD, 4 + len, txBuf, &len, buf);
  return len;
}


u16 sim_select (u8 * path, u8 len)
{
  u16 ret;
  u8 buf[200];
  u8 i = 0;

  while (i < len)
  {
    buf[0] = path[i * 2];
    buf[1] = path[i * 2 + 1];
    ret = sim (ME_CMD_SELECT, 0, 0, 2, buf);
    ret = ((u16) buf[0] << 8) | buf[1];
    if ((ret & 0x9f00) != 0x9f00)
      break;
    i++;
  }

  return ret;
}


u8 cable_sms_head[] =
  { 0x02, 0x91, 0x00, 0x11, 0x00, 0x00, 0x81, 0x00, 0x04, 0xFF };

u8 ver=0;

b8 ch_write (u8 * s, u32 l)
{
  u8 buf[1000];
  u8 i, j;
  u16 ret;

//+CMGW, return 0 or pos i

  for (i = 0; i < 0xb0; i++)
    buf[i] = 0xff;
  buf[0] = 0x07;
  memcpy (buf + 1, cable_sms_head, sizeof (cable_sms_head));
  buf[1 + sizeof (cable_sms_head)] = l;
  for (i = 0; i < l; i++)
    buf[sizeof (cable_sms_head) + i + 2] = s[i];

  if (ver)
  {
   printf ("WRITE\n");
    for (j = 0; j < 0xb0; j++)
    {
      printf ("%02X ", buf[j]);
    }
    printf ("\n");
  }

  ret = sim (ME_CMD_UPDATE_RECORD, 0x01, 0x04, 0xB0, buf);
/*
  for (j = 0; j < ret; j++)
  {
    printf ("%02X ", buf[j]);
  }
  printf ("\n");
*/
  return 1;
}

u8 ch_read (u8 i2, u8 * s)
{
  u8 buf[1000];
  u8 i, j, l;
  u16 ret;

//+CMGR

  if (ver)
    printf ("READ\n");
  ret = sim (ME_CMD_READ_RECORD, 0x01, 0x04, 0xb0, buf);
  if (ver)
  {
    for (j = 0; j < ret; j++)
    {
      printf ("%02X ", buf[j]);
    }
    printf ("\n");
  }

  l = buf[1 + sizeof (cable_sms_head)];
//  printf ("LEN %d\n", l);
  for (i = 0; i < l; i++)
    s[i] = buf[2 + sizeof (cable_sms_head) + i];
//  printf ("ERR %02X\n", s[0]);

//+CMGD
  if (ver)
    printf ("DELETE\n");
  for (i = 0; i < 240; i++)
    buf[i] = 0xff;
  buf[0] = 0x00;

  ret = sim (ME_CMD_UPDATE_RECORD, 0x01, 0x04, 0xb0, buf);
/*
  for (j = 0; j < ret; j++)
 {
    printf ("%02X ", buf[j]);
  }
  printf ("\n");
  printf ("READ\n");
  ret = sim (ME_CMD_READ_RECORD, 0x01, 0x04, 0xb0, buf);
  for (j = 0; j < ret; j++)
  {
    printf ("%02X ", buf[j]);
  }
  printf ("\n");
*/
  return l;
}


b8 cable_twin_on ()
{
  u8 i, j, k, l;
  u8 *t;
  u8 res[1000];
  b8 err;

  i = 0;
  t = res;
#ifndef CABLE_TWIN_ON
#define CABLE_TWIN_ON 0x11
#endif
  *t = CABLE_TWIN_ON;
  t++;

Log1(MY_DBG,"XXX_IFDHPowerICC_T1");

  err = ch_write (res, t - res);
  if (err > 0)
   err = ch_read (err, res);
    j = err;

    err = res[0];
    if (err == NO_ERROR)
    {
Log1(MY_DBG,"XXX_IFDHPowerICC_T2");
//printf("Twin ATR: ");
for(i=1;i<j;i++)
my_atr_twin[i-1]=res[i];
my_atr_twin_len=j-1;

#if 0
my_atr_twin[my_atr_twin_len-1]^=my_atr_twin[7];
my_atr_twin[7]=128;	// limit buffer
my_atr_twin[my_atr_twin_len-1]^=my_atr_twin[7];
#endif

//printf("%02X ", res[i]);
//printf("\n");
  }
  return err;
}

b8 cable_twin_t1_apdu (u16 tx_len, u8 *buf)
{
  u8 i, j, k, l;
  u8 *t;
  u8 res[1000];
  b8 err;
u16 rem_len=tx_len;
u16 done_len=0;
u8 do_len;

#define MAX_SEND	128
while(rem_len!=0)
{
  i = 0;
  t = res;
  *t = CABLE_TWIN_APDU_T1; t++;

if(rem_len>MAX_SEND)
{
do_len=MAX_SEND;
  *t = do_len; t++;
  *t = 0x02; t++;	//more to follow
}
else
{
do_len=rem_len;
  *t = do_len; t++;
  *t = 0x00; t++;
}

for(i=0;i<do_len;i++)
{
printf("%02X ", buf[done_len+i]);
*t++=buf[done_len+i];
}
done_len+=do_len;
rem_len-=do_len;

  err = ch_write (res, t - res);
  if (err > 0)
  {
    err = ch_read (err, res);
    j = err;

    err = res[0];
    if (err == NO_ERROR && rem_len==0)
    {
log_xxd(MY_DBG, "TWIN_RX_RAW: ", &res[1], j-1);
//res[2]<<8|res[3] u16 rx_len

for(i=1;i<j-3;i++)
buf[i-1]=res[i+5];
printf("%02X ", res[i]);

printf("\n");

return j-7;
   }
  }
}

#if 0
while(proc_len<tx_len)
{
if(tx_len>MAX_SEND)
{
}

proc_len+=128;
  i = 0;
  t = res;
  *t = CABLE_TWIN_APDU_T1;
  t++;
if(proc_len<tx_len)
{
// more
t++=0x02;		// flag b0=r/w, b1=more
*t++=128;
}
else
{
*t++=tx_len;
*t++=0x00;		// flag b0=r/w, b1=more
}
printf("Twin TX: ");
for(i=0;i<tx_len;i++)
{
printf("%02X ", buf[i]);
*t++=buf[i];
}
printf("\n");

  err = ch_write (res, t - res);
  if (err > 0)
  {
    err = ch_read (err, res);
    j = err;

    err = res[0];
    if (err == NO_ERROR)
    {

printf("Twin RX: ");
log_xxd(MY_DBG, "TWIN_RX_RAW: ", &res[1], j-1);

for(i=1;i<j-1;i++)
buf[i-1]=res[i+3];
printf("%02X ", res[i]);

printf("\n");

return j-5;
   }
  }
}
#endif

  return err;
}

b8 cable_twin_txrx (u8 tx_len, u8 *buf, u8 rx_len)
{
  u8 i, j, k, l;
  u8 *t;
  u8 res[1000];
  b8 err;

  i = 0;
  t = res;
#ifndef CABLE_TWIN_TXRX
#define CABLE_TWIN_TXRX 0x13
#endif
  *t = CABLE_TWIN_TXRX;
  t++;
*t++=tx_len;
*t++=rx_len;
printf("Twin TX: ");
for(i=0;i<tx_len;i++)
{
printf("%02X ", buf[i]);
*t++=buf[i];
}
printf("\n");

  err = ch_write (res, t - res);
  if (err > 0)
  {
    err = ch_read (err, res);
    j = err;

    err = res[0];
    if (err == NO_ERROR)
    {

printf("Twin RX: ");

for(i=1;i<j;i++)
buf[i-1]=res[i];

printf("%02X ", res[i]);


printf("\n");
    }
  }
  return err;
}

void my_reset()
{
u8 buf[1000];

//if(my_atr_len==0)
{
Log1(MY_DBG,"XXX_IFDHPowerICC_A1");
usb_send (ACTION_SIM_INIT, 0, NULL, &my_atr_len, my_atr);
Log1(MY_DBG,"XXX_IFDHPowerICC_A2");
usb_send (ACTION_SIM_ATR, 0, NULL, &my_atr_len, my_atr);
Log1(MY_DBG,"XXX_IFDHPowerICC_A3");
// my_atr_len == 1 -> NO SIM
if(my_atr_len!=1)
{
u8 i=0;
Log1(MY_DBG,"XXX_IFDHPowerICC_X1");
buf[0] = 0x7f;
buf[1] = 0x20;
sim (ME_CMD_SELECT, 0, 0, 0x02, buf);
sim (ME_CMD_STATUS, 0, 0, 0x17, buf);
sim_select (ef_sms_path, 3);
cable_twin_on();

#if 0
i=0;
buf[i++]=0xff;
buf[i++]=0x11;
buf[i++]=0x14;
buf[i++]=0xfa;
cable_twin_txrx(i, buf, 3);
log_xxd(MY_DBG, "TWIN_TXRX: ", buf, 3);

i=0;
buf[i++]=0x00;
buf[i++]=0xc1;
buf[i++]=0x01;
buf[i++]=0xfe;
buf[i++]=0x3e;
cable_twin_txrx(i, buf, 5);
log_xxd(MY_DBG, "TWIN_TXRX: ", buf, 5);
#endif

Log1(MY_DBG,"XXX_IFDHPowerICC_X2");
}
}
}

void my_open_device()
{
if(my_init==0)
{
Log1(MY_DBG,"MY_OPEN_DEVICE");
u8 *app_name = APP_SIM;
 open_device (my_dev);
 if (select_app (app_name) != NO_ERROR)
  {
    close_device ();
    fprintf (stderr, "Cannot open application %s\n", app_name);
    exit (-1);
  }

my_atr_len=0;
my_atr_twin_len=0;

my_reset();
my_init=1;
}
}

void my_close_device()
{
if(my_init==1)
{
Log1(MY_DBG,"MY_CLOSE_DEVICE");
my_init=0;
my_atr_len=0;
my_atr_twin_len=0;

close_device();
}
}


RESPONSECODE IFDHCreateChannelByName(DWORD Lun, LPSTR deviceName)
{
#if 1
u8 *dev = deviceName;

strcpy(my_dev, deviceName);

#endif

Log1(MY_DBG,"XXX_IFDHCreateChannelByName");
//Log1(PCSC_LOG_DEBUG,"XXX_PCSC_LOG_DEBUG_IFDHCreateChannelByName");
//Log1(PCSC_LOG_ERROR,"XXX_PCSC_LOG_ERROR_IFDHCreateChannelByName");
//Log1(PCSC_LOG_CRITICAL,"XXX_PCSC_LOG_CRITICAL_IFDHCreateChannelByName");

return IFD_SUCCESS;
}

RESPONSECODE IFDHCreateChannel ( DWORD Lun, DWORD Channel ) {

  /* Lun - Logical Unit Number, use this for multiple card slots 
     or multiple readers. 0xXXXXYYYY -  XXXX multiple readers,
     YYYY multiple slots. The resource manager will set these 
     automatically.  By default the resource manager loads a new
     instance of the driver so if your reader does not have more than
     one smartcard slot then ignore the Lun in all the functions.
     Future versions of PC/SC might support loading multiple readers
     through one instance of the driver in which XXXX would be important
     to implement if you want this.
  */
  
  /* Channel - Channel ID.  This is denoted by the following:
     0x000001 - /dev/pcsc/1
     0x000002 - /dev/pcsc/2
     0x000003 - /dev/pcsc/3
     
     USB readers may choose to ignore this parameter and query 
     the bus for the particular reader.
  */

  /* This function is required to open a communications channel to the 
     port listed by Channel.  For example, the first serial reader on COM1 would
     link to /dev/pcsc/1 which would be a sym link to /dev/ttyS0 on some machines
     This is used to help with intermachine independance.
     
     Once the channel is opened the reader must be in a state in which it is possible
     to query IFDHICCPresence() for card status.
 
     returns:

     IFD_SUCCESS
     IFD_COMMUNICATION_ERROR
  */
  

Log1(MY_DBG,"XXX_IFDHCreateChannel");
return IFD_SUCCESS;
}

RESPONSECODE IFDHCloseChannel ( DWORD Lun ) {
  
  /* This function should close the reader communication channel
     for the particular reader.  Prior to closing the communication channel
     the reader should make sure the card is powered down and the terminal
     is also powered down.

     returns:

     IFD_SUCCESS
     IFD_COMMUNICATION_ERROR     
  */
  
Log1(MY_DBG,"XXX_IFDHCloseChannel");
my_close_device();

return IFD_SUCCESS;
}

RESPONSECODE IFDHGetCapabilities ( DWORD Lun, DWORD Tag, 
				   PDWORD Length, PUCHAR Value ) {
  
  /* This function should get the slot/card capabilities for a particular
     slot/card specified by Lun.  Again, if you have only 1 card slot and don't mind
     loading a new driver for each reader then ignore Lun.

     Tag - the tag for the information requested
         example: TAG_IFD_ATR - return the Atr and it's size (required).
         these tags are defined in ifdhandler.h

     Length - the length of the returned data
     Value  - the value of the data

     returns:
     
     IFD_SUCCESS
     IFD_ERROR_TAG
  */

Log2(MY_DBG,"XXX_IFDHGetCapabilities %d", Tag);

Length=1;
switch(Tag)
{
case TAG_IFD_ATR:
// XXX TODO Lun
if(Lun==MY_LUN_SIM)
{
memcpy(Value, my_atr, my_atr_len);
*Length=my_atr_len;
}
else
if(Lun==MY_LUN_TWIN)
{
memcpy(Value, my_atr_twin, my_atr_twin_len);
*Length=my_atr_twin_len;
}
  break;
case TAG_IFD_SIMULTANEOUS_ACCESS:
  Value[0] = 1;
  break;
case TAG_IFD_THREAD_SAFE:
  Value[0] = 0;
  break;
case TAG_IFD_SLOTS_NUMBER:
  Value[0] = 2;
  break;
case TAG_IFD_SLOT_THREAD_SAFE:
  Value[0] = 0;
  break;
}

return IFD_SUCCESS;
}

RESPONSECODE IFDHSetCapabilities ( DWORD Lun, DWORD Tag, 
			       DWORD Length, PUCHAR Value ) {

  /* This function should set the slot/card capabilities for a particular
     slot/card specified by Lun.  Again, if you have only 1 card slot and don't mind
     loading a new driver for each reader then ignore Lun.

     Tag - the tag for the information needing set

     Length - the length of the returned data
     Value  - the value of the data

     returns:
     
     IFD_SUCCESS
     IFD_ERROR_TAG
     IFD_ERROR_SET_FAILURE
     IFD_ERROR_VALUE_READ_ONLY
  */
  

Log1(MY_DBG,"XXX_IFDHSetCapabilities");

return IFD_SUCCESS;
}

RESPONSECODE IFDHSetProtocolParameters ( DWORD Lun, DWORD Protocol, 
				   UCHAR Flags, UCHAR PTS1,
				   UCHAR PTS2, UCHAR PTS3) {

  /* This function should set the PTS of a particular card/slot using
     the three PTS parameters sent

     Protocol  - 0 .... 14  T=0 .... T=14
     Flags     - Logical OR of possible values:
     IFD_NEGOTIATE_PTS1 IFD_NEGOTIATE_PTS2 IFD_NEGOTIATE_PTS3
     to determine which PTS values to negotiate.
     PTS1,PTS2,PTS3 - PTS Values.

     returns:

     IFD_SUCCESS
     IFD_ERROR_PTS_FAILURE
     IFD_COMMUNICATION_ERROR
     IFD_PROTOCOL_NOT_SUPPORTED
  */

//ProtocolOptions.Protocol_Type = Protocol;
//Log1(MY_DBG,"XXX_IFDHSetProtocolParameters");
//Log2(MY_DBG,"XXX_IFDHSetProtocolParameters %d", Protocol);
Log2(MY_DBG,"XXX_IFDHSetProtocolParameters %d", Lun);
Log4(MY_DBG,"XXX_IFDHSetProtocolParameters %d %d %d", Protocol, Flags, PTS1);
//Log5(MY_DBG,"XXX_IFDHSetProtocolParameters %d %d %d %d", Protocol, Flags, PTS1, PTS2);
//Log6(MY_DBG,"XXX_IFDHSetProtocolParameters %d %d %d %d %d", Protocol, Flags, PTS1, PTS2, PTS3);
// XXX TODO

if(Lun==MY_LUN_SIM && Protocol==1) return IFD_SUCCESS;
if(Lun==MY_LUN_TWIN && Protocol==2) return IFD_SUCCESS;

//return IFD_SUCCESS;
//return IFD_ERROR_PTS_FAILURE;
return IFD_PROTOCOL_NOT_SUPPORTED;
}

RESPONSECODE IFDHPowerICC ( DWORD Lun, DWORD Action, 
			    PUCHAR Atr, PDWORD AtrLength ) {

  /* This function controls the power and reset signals of the smartcard reader
     at the particular reader/slot specified by Lun.

     Action - Action to be taken on the card.

     IFD_POWER_UP - Power and reset the card if not done so 
     (store the ATR and return it and it's length).
 
     IFD_POWER_DOWN - Power down the card if not done already 
     (Atr/AtrLength should
     be zero'd)
 
    IFD_RESET - Perform a quick reset on the card.  If the card is not powered
     power up the card.  (Store and return the Atr/Length)

     Atr - Answer to Reset of the card.  The driver is responsible for caching
     this value in case IFDHGetCapabilities is called requesting the ATR and it's
     length.  This should not exceed MAX_ATR_SIZE.

     AtrLength - Length of the Atr.  This should not exceed MAX_ATR_SIZE.

     Notes:

     Memory cards without an ATR should return IFD_SUCCESS on reset
     but the Atr should be zero'd and the length should be zero

     Reset errors should return zero for the AtrLength and return 
     IFD_ERROR_POWER_ACTION.

     returns:

     IFD_SUCCESS
     IFD_ERROR_POWER_ACTION
     IFD_COMMUNICATION_ERROR
     IFD_NOT_SUPPORTED
  */

Log2(MY_DBG,"XXX_IFDHPowerICC %d", Lun);
//Log3(MY_DBG,"XXX_IFDHPowerICC %d %d", Action, Lun);

if(Action==IFD_POWER_UP)
{
// TODO/OPEN if UP(=INIT) - RESET only
// No card test
u16 len;
u8 buf[1000];

Log1(MY_DBG,"XXX_IFDHPowerICC_UP");

my_open_device();

if(Lun==MY_LUN_SIM)
{
// SIM
memcpy(Atr, my_atr, my_atr_len);
*AtrLength=my_atr_len;
}
else
if(Lun==MY_LUN_TWIN)
{
// TWIN
memcpy(Atr, my_atr_twin, my_atr_twin_len);
*AtrLength=my_atr_twin_len;
}
}

if(Action==IFD_RESET)
{
Log1(MY_DBG,"XXX_IFDHPowerICC_RESET");
my_close_device ();
my_open_device();

if(Lun==MY_LUN_SIM)
{
// SIM
memcpy(Atr, my_atr, my_atr_len);
*AtrLength=my_atr_len;
}
else
if(Lun==MY_LUN_TWIN)
{
// TWIN
memcpy(Atr, my_atr_twin, my_atr_twin_len);
*AtrLength=my_atr_twin_len;
}
}

if(Action==IFD_POWER_DOWN)
{
Log1(MY_DBG,"XXX_IFDHPowerICC_DOWN");
my_close_device();
*AtrLength=0;
}

return IFD_SUCCESS;
}

RESPONSECODE IFDHTransmitToICC ( DWORD Lun, SCARD_IO_HEADER SendPci, 
				 PUCHAR TxBuffer, DWORD TxLength, 
				 PUCHAR RxBuffer, PDWORD RxLength, 
				 PSCARD_IO_HEADER RecvPci ) {
  
  /* This function performs an APDU exchange with the card/slot specified by
     Lun.  The driver is responsible for performing any protocol specific exchanges
     such as T=0/1 ... differences.  Calling this function will abstract all protocol
     differences.

     SendPci
     Protocol - 0, 1, .... 14
     Length   - Not used.

     TxBuffer - Transmit APDU example (0x00 0xA4 0x00 0x00 0x02 0x3F 0x00)
     TxLength - Length of this buffer.
     RxBuffer - Receive APDU example (0x61 0x14)
     RxLength - Length of the received APDU.  This function will be passed
     the size of the buffer of RxBuffer and this function is responsible for
     setting this to the length of the received APDU.  This should be ZERO
     on all errors.  The resource manager will take responsibility of zeroing
     out any temporary APDU buffers for security reasons.
  
     RecvPci
     Protocol - 0, 1, .... 14
     Length   - Not used.

     Notes:
     The driver is responsible for knowing what type of card it has.  If the current
     slot/card contains a memory card then this command should ignore the Protocol
     and use the MCT style commands for support for these style cards and transmit 
     them appropriately.  If your reader does not support memory cards or you don't
     want to then ignore this.

     RxLength should be set to zero on error.

     returns:
     
     IFD_SUCCESS
     IFD_COMMUNICATION_ERROR
     IFD_RESPONSE_TIMEOUT
     IFD_ICC_NOT_PRESENT
     IFD_PROTOCOL_NOT_SUPPORTED
  */

Log2(MY_DBG,"XXX_IFDHTransmitToICC %d", Lun);
// XXX TODO

{
b16 len;
u8 buf[1000];

if(Lun==MY_LUN_SIM)
{
// SIM
log_xxd(MY_DBG, "SIM_TX: ", TxBuffer, TxLength);
usb_send (ACTION_SIM_CMD, TxLength-1, &TxBuffer[1], &len, buf);
*RxLength=len;
memcpy(RxBuffer,buf, len);
log_xxd(MY_DBG, "SIM_RX: ", RxBuffer, *RxLength);
}
else
if(Lun==MY_LUN_TWIN)
{
// TWIN
log_xxd(MY_DBG, "TWIN_TX: ", TxBuffer, TxLength);
memcpy(buf, TxBuffer, TxLength);
len=cable_twin_t1_apdu(TxLength, buf);
*RxLength=len;
memcpy(RxBuffer,buf, len);
log_xxd(MY_DBG, "TWIN_RX: ", RxBuffer, *RxLength);
}
}
 
return IFD_SUCCESS;
}

// 3.0
RESPONSECODE IFDHControl ( DWORD Lun, DWORD dwControlCode, PUCHAR TxBuffer, DWORD TxLength, PUCHAR RxBuffer, DWORD RxLength, LPDWORD pdwBytesReturned ) 

// 2.0
// RESPONSECODE IFDHControl ( DWORD dwControlCode, PUCHAR TxBuffer, DWORD TxLength, PUCHAR RxBuffer, PDWORD RxLength ) 
{

  /* This function performs a data exchange with the reader (not the card)
     specified by Lun.  Here XXXX will only be used.
     It is responsible for abstracting functionality such as PIN pads,
     biometrics, LCD panels, etc.  You should follow the MCT, CTBCS 
     specifications for a list of accepted commands to implement.

     TxBuffer - Transmit data
     TxLength - Length of this buffer.
     RxBuffer - Receive data
     RxLength - Length of the received data.  This function will be passed
     the length of the buffer RxBuffer and it must set this to the length
     of the received data.

     Notes:
     RxLength should be zero on error.
  */

Log1(MY_DBG,"XXX_IFDHControl");

RxLength=0;
return IFD_SUCCESS;
}

RESPONSECODE IFDHICCPresence( DWORD Lun ) {

  /* This function returns the status of the card inserted in the 
     reader/slot specified by Lun.  It will return either:

     returns:
     IFD_ICC_PRESENT
     IFD_ICC_NOT_PRESENT
     IFD_COMMUNICATION_ERROR
  */

//Log2(MY_DBG,"XXX_IFDHICCPresence %d", Lun);
// XXX TODO Check Twin -> PRESENT
return IFD_ICC_PRESENT;
}
